# -*- coding: utf-8 -*-
"""otr - WeeChat script for Off-the-Record IRC messaging

DISCLAIMER: To the best of my knowledge this script securely provides OTR
messaging in WeeChat, but I offer no guarantee. Please report any security
holes you find.

Copyright (c) 2012-2015 Matthew M. Boedicker <matthewm@boedicker.org>
                        Nils Görs <weechatter@arcor.de>
                        Daniel "koolfy" Faucon <koolfy@koolfy.be>
                        Felix Eckhofer <felix@tribut.de>

Report issues at https://github.com/mmb/weechat-otr

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# pylint: disable=too-many-lines

from __future__ import unicode_literals

import collections
import glob
import io
import os
import platform
import re
import traceback
import shlex
import shutil
import sys

import potr
import weechat

class PythonVersion2(object):
    """Python 2 version of code that must differ between Python 2 and 3."""

    def __init__(self):
        import cgi
        self.cgi = cgi

        import HTMLParser
        self.html_parser = HTMLParser
        self.html_parser_init_kwargs = {}

        import htmlentitydefs
        self.html_entities = htmlentitydefs

    def html_escape(self, strng):
        """Escape HTML characters in a string."""
        return self.cgi.escape(strng)

    def unicode(self, *args, **kwargs):
        """Return the Unicode version of a string."""
        return unicode(*args, **kwargs)

    def unichr(self, *args, **kwargs):
        """Return the one character string of a Unicode character number."""
        return unichr(*args, **kwargs)

    def to_unicode(self, strng):
        """Convert a utf-8 encoded string to a Unicode."""
        if isinstance(strng, unicode):
            return strng
        return strng.decode('utf-8', 'replace')

    def to_str(self, strng):
        """Convert a Unicode to a utf-8 encoded string."""
        return strng.encode('utf-8', 'replace')

class PythonVersion3(object):
    """Python 3 version of code that must differ between Python 2 and 3."""

    def __init__(self, minor):
        self.minor = minor

        import html
        self.html = html

        import html.parser
        self.html_parser = html.parser
        if self.minor >= 4:
            self.html_parser_init_kwargs = {'convert_charrefs' : True}
        else:
            self.html_parser_init_kwargs = {}

        import html.entities
        self.html_entities = html.entities

    def html_escape(self, strng):
        """Escape HTML characters in a string."""
        return self.html.escape(strng, quote=False)

    def unicode(self, *args, **kwargs):
        """Return the Unicode version of a string."""
        return str(*args, **kwargs)

    def unichr(self, *args, **kwargs):
        """Return the one character string of a Unicode character number."""
        return chr(*args, **kwargs)

    def to_unicode(self, strng):
        """Convert a utf-8 encoded string to unicode."""
        if isinstance(strng, bytes):
            return strng.decode('utf-8', 'replace')
        return strng

    def to_str(self, strng):
        """Convert a Unicode to a utf-8 encoded string."""
        return strng

# We cannot use version_info.major as this is only supported on python >= 2.7
if sys.version_info[0] >= 3:
    PYVER = PythonVersion3(sys.version_info.minor)
else:
    PYVER = PythonVersion2()

SCRIPT_NAME = 'otr'
SCRIPT_DESC = 'Off-the-Record messaging for IRC'
SCRIPT_HELP = """{description}

Quick start:

Add an OTR item to the status bar by adding '[otr]' to the config setting
weechat.bar.status.items. This will show you whether your current conversation
is encrypted, authenticated and logged. /set otr.* for OTR status bar
customization options.

Start a private conversation with a friend who has OTR: /query yourpeer hi

In the private chat buffer: /otr start

If you have not authenticated your peer yet, follow the instructions for
authentication.

You can, at any time, see the current OTR session status and fingerprints with:
/otr status

View OTR policies for your peer: /otr policy

View default OTR policies: /otr policy default

Start/Stop log recording for the current OTR session: /otr log [start|stop]
This will be reverted back to the previous log setting at the end of the session.

To refresh the OTR session: /otr refresh

To end your private conversation: /otr finish

This script supports only OTR protocol version 2.
""".format(description=SCRIPT_DESC)

SCRIPT_AUTHOR = 'Matthew M. Boedicker'
SCRIPT_LICENCE = 'GPL3'
SCRIPT_VERSION = '1.9.2'

OTR_DIR_NAME = 'otr'

OTR_QUERY_RE = re.compile(r'\?OTR(\?|\??v[a-z\d]*\?)')

POLICIES = {
    'allow_v2' : 'allow OTR protocol version 2, effectively enable OTR '
                 'since v2 is the only supported version',
    'require_encryption' : 'refuse to send unencrypted messages when OTR is '
                           'enabled',
    'log' : 'enable logging of OTR conversations',
    'send_tag' : 'advertise your OTR capability using the whitespace tag',
    'html_escape' : 'escape HTML special characters in outbound messages',
    'html_filter' : 'filter HTML in incoming messages',
    }

READ_ONLY_POLICIES = {
    'allow_v1' : False,
    }

ACTION_PREFIX = '/me '
IRC_ACTION_RE = re.compile('^\x01ACTION (?P<text>.*)\x01$')
PLAIN_ACTION_RE = re.compile('^'+ACTION_PREFIX+'(?P<text>.*)$')

IRC_SANITIZE_TABLE = dict((ord(char), None) for char in '\n\r\x00')

# Patch potr.proto.TaggedPlaintext to not end plaintext tags in a space.
#
# When POTR adds OTR tags to plaintext it puts them at the end of the message.
# The tags end in a space which gets stripped off by WeeChat because it
# strips trailing spaces from commands. This causes OTR initiation to fail so
# the following code adds an extra tab at the end of the plaintext tags if
# they end in a space.
#
# The patched version also skips OTR tagging for CTCP messages because it
# breaks the CTCP format.
def patched__bytes__(self):
    """Patched potr.proto.TaggedPlainText.__bytes__."""
    # Do not tag CTCP messages.
    if self.msg.startswith(b'\x01') and \
        self.msg.endswith(b'\x01'):
        return self.msg

    data = self.msg + potr.proto.MESSAGE_TAG_BASE
    for v in self.versions:
        data += potr.proto.MESSAGE_TAGS[v]
    if data.endswith(b' '):
        data += b'\t'
    return data

potr.proto.TaggedPlaintext.__bytes__ = patched__bytes__

def command(buf, command_str):
    """Wrap weechat.command() with utf-8 encode."""
    debug(command_str)
    weechat.command(buf, PYVER.to_str(command_str))

def privmsg(server, nick, message):
    """Send a private message to a nick."""
    for line in message.splitlines():
        command('', '/quote -server {server} PRIVMSG {nick} :{line}'.format(
            server=irc_sanitize(server),
            nick=irc_sanitize(nick),
            line=irc_sanitize(line)))

def build_privmsg_in(fromm, target, msg):
    """Build inbound IRC PRIVMSG command."""
    return ':{user} PRIVMSG {target} :{msg}'.format(
        user=irc_sanitize(fromm),
        target=irc_sanitize(target),
        msg=irc_sanitize(msg))

def build_privmsgs_in(fromm, target, msg, prefix=''):
    """Build an inbound IRC PRIVMSG command for each line in msg.
    If prefix is supplied, prefix each line of msg with it."""
    cmd = []
    for line in msg.splitlines():
        cmd.append(build_privmsg_in(fromm, target, prefix+line))
    return '\r\n'.join(cmd)

def build_privmsg_out(target, msg):
    """Build outbound IRC PRIVMSG command(s)."""
    cmd = []
    for line in msg.splitlines():
        cmd.append('PRIVMSG {target} :{line}'.format(
            target=irc_sanitize(target),
            line=irc_sanitize(line)))
    return '\r\n'.join(cmd)

def irc_sanitize(msg):
    """Remove NUL, CR and LF characters from msg.
    The (utf-8 encoded version of a) string returned from this function
    should be safe to use as an argument in an irc command."""
    return PYVER.unicode(msg).translate(IRC_SANITIZE_TABLE)

def prnt(buf, message):
    """Wrap weechat.prnt() with utf-8 encode."""
    weechat.prnt(buf, PYVER.to_str(message))

def print_buffer(buf, message, level='info'):
    """Print message to buf with prefix,
    using color according to level."""
    prnt(buf, '{prefix}\t{msg}'.format(
        prefix=get_prefix(),
        msg=colorize(message, 'buffer.{0}'.format(level))))

def get_prefix():
    """Returns configured message prefix."""
    return weechat.string_eval_expression(
        config_string('look.prefix'),
        {}, {}, {})

def debug(msg):
    """Send a debug message to the OTR debug buffer."""
    debug_option = config_get_prefixed('general.debug')

    if not weechat.config_boolean(debug_option):
        return

    debug_buffer = weechat.buffer_search('python', 'OTR Debug')
    if not debug_buffer:
        debug_buffer = weechat.buffer_new('OTR Debug', '', '', '', '')
        weechat.buffer_set(debug_buffer, 'title', 'OTR Debug')
        weechat.buffer_set(debug_buffer, 'localvar_set_no_log', '1')

    prnt(debug_buffer, ('{script} debug\t{text}'.format(
        script=SCRIPT_NAME, text=PYVER.unicode(msg))))

def current_user(server_name):
    """Get the nick and server of the current user on a server."""
    return irc_user(info_get('irc_nick', server_name), server_name)

def irc_user(nick, server):
    """Build an IRC user string from a nick and server."""
    return '{nick}@{server}'.format(
        nick=nick.lower(),
        server=server)

def isupport_value(server, feature):
    """Get the value of an IRC server feature."""
    args = '{server},{feature}'.format(server=server, feature=feature)
    return info_get('irc_server_isupport_value', args)

def is_a_channel(channel, server):
    """Return true if a string has an IRC channel prefix."""
    prefixes = \
        tuple(isupport_value(server, 'CHANTYPES')) + \
        tuple(isupport_value(server, 'STATUSMSG'))

    # If the server returns nothing for CHANTYPES and STATUSMSG use
    # default prefixes.
    if not prefixes:
        prefixes = ('#', '&', '+', '!', '@')

    return channel.startswith(prefixes)

class PrivmsgParseException(Exception):
    """Exception class for PRIVMSG parsing exceptions."""
    pass

def parse_irc_privmsg(message, server):
    """Parse an IRC PRIVMSG command and return a dictionary.

    Either the to_channel key or the to_nick key will be set depending on
    whether the message is to a nick or a channel. The other will be None.

    Example input:

    :nick!user@host PRIVMSG #weechat :message here

    Output:

    {'from': 'nick!user@host',
    'from_nick': 'nick',
    'to': '#weechat',
    'to_channel': '#weechat',
    'to_nick': None,
    'text': 'message here'}
    """

    weechat_result = weechat.info_get_hashtable(
        'irc_message_parse', dict(message=message))

    if weechat_result['command'].upper() == 'PRIVMSG':
        target, text = PYVER.to_unicode(
            weechat_result['arguments']).split(' ', 1)
        if text.startswith(':'):
            text = text[1:]

        result = {
            'from': PYVER.to_unicode(weechat_result['host']),
            'to' : target,
            'text': text,
            }

        if weechat_result['host']:
            result['from_nick'] = PYVER.to_unicode(weechat_result['nick'])
        else:
            result['from_nick'] = ''

        if is_a_channel(target, server):
            result['to_channel'] = target
            result['to_nick'] = None
        else:
            result['to_channel'] = None
            result['to_nick'] = target

        return result
    else:
        raise PrivmsgParseException(message)

def has_otr_end(msg):
    """Return True if the message is the end of an OTR message."""
    return msg.endswith('.') or msg.endswith(',')

def first_instance(objs, klass):
    """Return the first object in the list that is an instance of a class."""
    for obj in objs:
        if isinstance(obj, klass):
            return obj

def config_prefix(option):
    """Add the config prefix to an option and return the full option name."""
    return '{script}.{option}'.format(
        script=SCRIPT_NAME,
        option=option)

def config_color(option):
    """Get the color of a color config option."""
    return weechat.color(weechat.config_color(config_get_prefixed(
        'color.{0}'.format(option))))

def config_string(option):
    """Get the string value of a config option with utf-8 decode."""
    return PYVER.to_unicode(weechat.config_string(
        config_get_prefixed(option)))

def config_get(option):
    """Get the value of a WeeChat config option."""
    return weechat.config_get(PYVER.to_str(option))

def config_get_prefixed(option):
    """Get the value of a script prefixed WeeChat config option."""
    return config_get(config_prefix(option))

def buffer_get_string(buf, prop):
    """Wrap weechat.buffer_get_string() with utf-8 encode/decode."""
    if buf is not None:
        encoded_buf = PYVER.to_str(buf)
    else:
        encoded_buf = None

    return PYVER.to_unicode(weechat.buffer_get_string(
        encoded_buf, PYVER.to_str(prop)))

def buffer_is_private(buf):
    """Return True if a buffer is private."""
    return buffer_get_string(buf, 'localvar_type') == 'private'

def info_get(info_name, arguments):
    """Wrap weechat.info_get() with utf-8 encode/decode."""
    return PYVER.to_unicode(weechat.info_get(
        PYVER.to_str(info_name), PYVER.to_str(arguments)))

def msg_irc_from_plain(msg):
    """Transform a plain-text message to irc format.
    This will replace lines that start with /me with the respective
    irc command."""
    return PLAIN_ACTION_RE.sub('\x01ACTION \g<text>\x01', msg)

def msg_plain_from_irc(msg):
    """Transform an irc message to plain-text.
    Any ACTION found will be rewritten as /me <text>."""
    return IRC_ACTION_RE.sub(ACTION_PREFIX + r'\g<text>', msg)

def default_peer_args(args, buf):
    """Get the nick and server of a remote peer from command arguments or
    a buffer.

    args is the [nick, server] slice of arguments from a command.
    If these are present, return them. If args is empty and the buffer buf
    is private, return the remote nick and server of buf."""
    result = None, None

    if len(args) == 2:
        result = tuple(args)
    else:
        if buffer_is_private(buf):
            result = (
                buffer_get_string(buf, 'localvar_channel'),
                buffer_get_string(buf, 'localvar_server'))

    return result

def format_default_policies():
    """Return current default policies formatted as a string for the user."""
    buf = io.StringIO()

    buf.write('Current default OTR policies:\n')

    for policy, desc in sorted(POLICIES.items()):
        buf.write('  {policy} ({desc}) : {value}\n'.format(
            policy=policy,
            desc=desc,
            value=config_string('policy.default.{0}'.format(policy))))

    buf.write('Change default policies with: /otr policy default NAME on|off')

    return buf.getvalue()

def to_bytes(strng):
    """Convert a python str or unicode to bytes."""
    return strng.encode('utf-8', 'replace')

def colorize(msg, color):
    """Colorize each line of msg using color."""
    result = []
    colorstr = config_color(color)

    for line in msg.splitlines():
        result.append('{color}{msg}'.format(
            color=colorstr,
            msg=line))

    return '\r\n'.join(result)

def accounts():
    """Return a list of all IrcOtrAccounts sorted by name."""
    result = []
    for key_path in glob.iglob(os.path.join(OTR_DIR, '*.key3')):
        key_name, _ = os.path.splitext(os.path.basename(key_path))
        result.append(ACCOUNTS[key_name])

    return sorted(result, key=lambda account: account.name)

def show_account_fingerprints():
    """Print all account names and their fingerprints to the core buffer."""
    table_formatter = TableFormatter()
    for account in accounts():
        table_formatter.add_row([
            account.name,
            str(account.getPrivkey())])
    print_buffer('', table_formatter.format())

def show_peer_fingerprints(grep=None):
    """Print peer names and their fingerprints to the core buffer.

    If grep is passed in, show all peer names containing that substring."""
    trust_descs = {
        '' : 'unverified',
        'smp' : 'SMP verified',
        'verified' : 'verified',
        }

    table_formatter = TableFormatter()
    for account in accounts():
        for peer, peer_data in sorted(account.trusts.items()):
            for fingerprint, trust in sorted(peer_data.items()):
                if grep is None or grep in peer:
                    table_formatter.add_row([
                        peer,
                        account.name,
                        potr.human_hash(fingerprint),
                        trust_descs[trust],
                    ])
    print_buffer('', table_formatter.format())

def private_key_file_path(account_name):
    """Return the private key file path for an account."""
    return os.path.join(OTR_DIR, '{0}.key3'.format(account_name))

def read_private_key(key_file_path):
    """Return the private key in a private key file."""
    debug(('read private key', key_file_path))

    with open(key_file_path, 'rb') as key_file:
        return potr.crypt.PK.parsePrivateKey(key_file.read())[0]

def get_context(account_name, context_name):
    """Return a context from an account."""
    return ACCOUNTS[account_name].getContext(context_name)

def get_server_context(server, peer_nick):
    """Return the context for the current server user and peer."""
    return get_context(current_user(server), irc_user(peer_nick, server))

class AccountDict(collections.defaultdict):
    """Dictionary that adds missing keys as IrcOtrAccount instances."""

    def __missing__(self, key):
        debug(('add account', key))
        self[key] = IrcOtrAccount(key)

        return self[key]

class Assembler(object):
    """Reassemble fragmented OTR messages.

    This does not deal with OTR fragmentation, which is handled by potr, but
    fragmentation of received OTR messages that are too large for IRC.
    """
    def __init__(self):
        self.clear()

    def add(self, data):
        """Add data to the buffer."""
        self.value += data

    def clear(self):
        """Empty the buffer."""
        self.value = ''

    def is_done(self):
        """Return True if the buffer is a complete message."""
        return self.is_query() or \
            not to_bytes(self.value).startswith(potr.proto.OTRTAG) or \
            has_otr_end(self.value)

    def get(self):
        """Return the current value of the buffer and empty it."""
        result = self.value
        self.clear()

        return result

    def is_query(self):
        """Return true if the buffer is an OTR query."""
        return OTR_QUERY_RE.search(self.value)

class IrcContext(potr.context.Context):
    """Context class for OTR over IRC."""

    def __init__(self, account, peername):
        super(IrcContext, self).__init__(account, peername)

        self.peer_nick, self.peer_server = peername.split('@', 1)
        self.in_assembler = Assembler()
        self.in_otr_message = False
        self.in_smp = False
        self.smp_question = False

    def policy_config_option(self, policy):
        """Get the option name of a policy option for this context."""
        return config_prefix('.'.join([
            'policy', self.peer_server, self.user.nick, self.peer_nick,
            policy.lower()]))

    def getPolicy(self, key):
        """Get the value of a policy option for this context."""
        key_lower = key.lower()

        if key_lower in READ_ONLY_POLICIES:
            result = READ_ONLY_POLICIES[key_lower]
        elif key_lower == 'send_tag' and self.no_send_tag():
            result = False
        else:
            option = config_get(self.policy_config_option(key))

            if option == '':
                option = config_get(self.user.policy_config_option(key))

            if option == '':
                option = config_get_prefixed('.'.join(
                    ['policy', self.peer_server, key_lower]))

            if option == '':
                option = config_get_prefixed(
                    'policy.default.{0}'.format(key_lower))

            result = bool(weechat.config_boolean(option))

        debug(('getPolicy', key, result))

        return result

    def inject(self, msg, appdata=None):
        """Send a message to the remote peer."""
        if isinstance(msg, potr.proto.OTRMessage):
            msg = PYVER.unicode(msg)
        else:
            msg = PYVER.to_unicode(msg)

        debug(('inject', msg, 'len {0}'.format(len(msg)), appdata))

        privmsg(self.peer_server, self.peer_nick, msg)

    def setState(self, newstate):
        """Handle state transition."""
        debug(('state', self.state, newstate))

        if self.is_encrypted():
            if newstate == potr.context.STATE_ENCRYPTED:
                self.print_buffer(
                    'Private conversation has been refreshed.', 'success')
            elif newstate == potr.context.STATE_FINISHED:
                self.print_buffer(
                    '{peer} has ended the private conversation. You should do '
                    'the same:\n/otr finish'.format(peer=self.peer_nick))
        elif newstate == potr.context.STATE_ENCRYPTED:
            # unencrypted => encrypted
            trust = self.getCurrentTrust()

            # Disable logging before any proof of OTR activity is generated.
            # This is necessary when the session is started automatically, and
            # not by /otr start.
            if not self.getPolicy('log'):
                self.previous_log_level = self.disable_logging()
            else:
                self.previous_log_level = self.get_log_level()
                if self.is_logged():
                    self.hint(
                        'You have enabled the recording to disk of OTR '
                        'conversations. By doing this you are potentially '
                        'putting yourself and your correspondent in danger. '
                        'Please consider disabling this policy with '
                        '"/otr policy default log off". To disable logging '
                        'for this OTR session, use "/otr log stop"')

            if trust is None:
                fpr = str(self.getCurrentKey())
                self.print_buffer('New fingerprint: {0}'.format(fpr), 'warning')
                self.setCurrentTrust('')

            if bool(trust):
                self.print_buffer(
                    'Authenticated secured OTR conversation started.',
                    'success')
            else:
                self.print_buffer(
                    'Unauthenticated secured OTR conversation started.',
                    'warning')
                self.hint(self.verify_instructions())

        if self.state != potr.context.STATE_PLAINTEXT and \
                newstate == potr.context.STATE_PLAINTEXT:
            self.print_buffer('Private conversation ended.')

            # If we altered the logging value, restore it.
            if self.previous_log_level is not None:
                self.restore_logging(self.previous_log_level)

        super(IrcContext, self).setState(newstate)

    def maxMessageSize(self, appdata=None):
        """Return the max message size for this context."""
        # remove 'PRIVMSG <nick> :' from max message size
        result = self.user.maxMessageSize - 10 - len(self.peer_nick)
        debug('max message size {0}'.format(result))

        return result

    def buffer(self):
        """Get the buffer for this context."""
        return info_get(
            'irc_buffer', '{server},{nick}'.format(
                server=self.peer_server,
                nick=self.peer_nick
                ))

    def print_buffer(self, msg, level='info'):
        """Print a message to the buffer for this context.
        level is used to colorize the message."""
        buf = self.buffer()

        # add [nick] prefix if we have only a server buffer for the query
        if self.peer_nick and not buffer_is_private(buf):
            msg = '[{nick}] {msg}'.format(
                nick=self.peer_nick,
                msg=msg)

        print_buffer(buf, msg, level)

    def hint(self, msg):
        """Print a message to the buffer but only when hints are enabled."""
        hints_option = config_get_prefixed('general.hints')

        if weechat.config_boolean(hints_option):
            self.print_buffer(msg, 'hint')

    def smp_finish(self, message=False, level='info'):
        """Reset SMP state and send a message to the user."""
        self.in_smp = False
        self.smp_question = False

        self.user.saveTrusts()
        if message:
            self.print_buffer(message, level)

    def handle_tlvs(self, tlvs):
        """Handle SMP states."""
        if tlvs:
            smp1q = first_instance(tlvs, potr.proto.SMP1QTLV)
            smp3 = first_instance(tlvs, potr.proto.SMP3TLV)
            smp4 = first_instance(tlvs, potr.proto.SMP4TLV)

            if first_instance(tlvs, potr.proto.SMPABORTTLV):
                debug('SMP aborted by peer')
                self.smp_finish('SMP aborted by peer.', 'warning')
            elif self.in_smp and not self.smpIsValid():
                debug('SMP aborted')
                self.smp_finish('SMP aborted.', 'error')
            elif first_instance(tlvs, potr.proto.SMP1TLV):
                debug('SMP1')
                self.in_smp = True

                self.print_buffer(
                    """Peer has requested SMP verification.
Respond with: /otr smp respond <secret>""")
            elif smp1q:
                debug(('SMP1Q', smp1q.msg))
                self.in_smp = True
                self.smp_question = True

                self.print_buffer(
                    """Peer has requested SMP verification: {msg}
Respond with: /otr smp respond <answer>""".format(
    msg=PYVER.to_unicode(smp1q.msg)))
            elif first_instance(tlvs, potr.proto.SMP2TLV):
                if not self.in_smp:
                    debug('Received unexpected SMP2')
                    self.smp_finish()
                else:
                    debug('SMP2')
                    self.print_buffer('SMP progressing.')
            elif smp3 or smp4:
                if smp3:
                    debug('SMP3')
                elif smp4:
                    debug('SMP4')

                if self.smpIsSuccess():

                    if self.smp_question:
                        self.smp_finish(
                            'SMP verification succeeded.', 'success')
                        if not self.is_verified:
                            self.print_buffer(
                                'You may want to authenticate your peer by '
                                'asking your own question:\n'
                                "/otr smp ask <'question'> 'secret'")
                    else:
                        self.smp_finish(
                            'SMP verification succeeded.', 'success')

                else:
                    self.smp_finish('SMP verification failed.', 'error')

    def verify_instructions(self):
        """Generate verification instructions for user."""
        return """You can verify that this contact is who they claim to be in
one of the following ways:

1) Verify each other's fingerprints using a secure channel:
  Your fingerprint : {your_fp}
  {peer}'s fingerprint : {peer_fp}
  then use the command: /otr trust {peer_nick} {peer_server}

2) SMP pre-shared secret that you both know:
  /otr smp ask {peer_nick} {peer_server} 'secret'

3) SMP pre-shared secret that you both know with a question:
  /otr smp ask {peer_nick} {peer_server} <'question'> 'secret'

Note: You can safely omit specifying the peer and server when
      executing these commands from the appropriate conversation
      buffer
""".format(
    your_fp=self.user.getPrivkey(),
    peer=self.peer,
    peer_nick=self.peer_nick,
    peer_server=self.peer_server,
    peer_fp=potr.human_hash(self.crypto.theirPubkey.cfingerprint()))

    def is_encrypted(self):
        """Return True if the conversation with this context's peer is
        currently encrypted."""
        return self.state == potr.context.STATE_ENCRYPTED

    def is_verified(self):
        """Return True if this context's peer is verified."""
        return bool(self.getCurrentTrust())

    def format_policies(self):
        """Return current policies for this context formatted as a string for
        the user."""
        buf = io.StringIO()

        buf.write('Current OTR policies for {peer}:\n'.format(
            peer=self.peer))

        for policy, desc in sorted(POLICIES.items()):
            buf.write('  {policy} ({desc}) : {value}\n'.format(
                policy=policy,
                desc=desc,
                value='on' if self.getPolicy(policy) else 'off'))

        buf.write('Change policies with: /otr policy NAME on|off')

        return buf.getvalue()

    def is_logged(self):
        """Return True if conversations with this context's peer are currently
        being logged to disk."""
        infolist = weechat.infolist_get('logger_buffer', '', '')

        buf = self.buffer()

        result = False

        while weechat.infolist_next(infolist):
            if weechat.infolist_pointer(infolist, 'buffer') == buf:
                result = bool(weechat.infolist_integer(infolist, 'log_enabled'))
                break

        weechat.infolist_free(infolist)

        return result

    def get_log_level(self):
        """Return the current logging level for this context's peer
        or -1 if the buffer uses the default log level of weechat."""
        infolist = weechat.infolist_get('logger_buffer', '', '')

        buf = self.buffer()

        if not config_get(self.get_logger_option_name()):
            result = -1
        else:
            result = 0

            while weechat.infolist_next(infolist):
                if weechat.infolist_pointer(infolist, 'buffer') == buf:
                    result = weechat.infolist_integer(infolist, 'log_level')
                    break

            weechat.infolist_free(infolist)

        return result

    def get_logger_option_name(self):
        """Returns the logger config option for this context's buffer."""
        buf = self.buffer()
        name = buffer_get_string(buf, 'name')
        plugin = buffer_get_string(buf, 'plugin')

        return 'logger.level.{plugin}.{name}'.format(
            plugin=plugin, name=name)

    def disable_logging(self):
        """Return the previous logger level and set the buffer logger level
        to 0. If it was already 0, return None."""
        # If previous_log_level has not been previously set, return the level
        # we detect now.
        if not hasattr(self, 'previous_log_level'):
            previous_log_level = self.get_log_level()

            if self.is_logged():
                weechat.command(self.buffer(), '/mute logger disable')
                self.print_buffer(
                    'Logs have been temporarily disabled for the session. '
                    'They will be restored upon finishing the OTR session.')

            return previous_log_level

        # If previous_log_level was already set, it means we already altered it
        # and that we just detected an already modified logging level.
        # Return the pre-existing value so it doesn't get lost, and we can
        # restore it later.
        else:
            return self.previous_log_level

    def restore_logging(self, previous_log_level):
        """Restore the log level of the buffer."""
        buf = self.buffer()

        if (previous_log_level >= 0) and (previous_log_level < 10):
            self.print_buffer(
                'Restoring buffer logging value to: {0}'.format(
                    previous_log_level), 'warning')
            weechat.command(buf, '/mute logger set {0}'.format(
                previous_log_level))

        if previous_log_level == -1:
            logger_option_name = self.get_logger_option_name()
            self.print_buffer(
                'Restoring buffer logging value to default', 'warning')
            weechat.command(buf, '/mute unset {0}'.format(
                logger_option_name))

        del self.previous_log_level

    def msg_convert_in(self, msg):
        """Transform incoming OTR message to IRC format.
        This includes stripping html, converting plain-text ACTIONs
        and character encoding conversion.
        Only character encoding is changed if context is unencrypted."""
        msg = PYVER.to_unicode(msg)

        if not self.is_encrypted():
            return msg

        if self.getPolicy('html_filter'):
            try:
                msg = IrcHTMLParser.parse(msg)
            except PYVER.html_parser.HTMLParseError:
                pass

        return msg_irc_from_plain(msg)

    def msg_convert_out(self, msg):
        """Convert an outgoing IRC message to be sent over OTR.
        This includes escaping html, converting ACTIONs to plain-text
        and character encoding conversion
        Only character encoding is changed if context is unencrypted."""
        if self.is_encrypted():
            msg = msg_plain_from_irc(msg)

            if self.getPolicy('html_escape'):
                msg = PYVER.html_escape(msg)

        # potr expects bytes to be returned
        return to_bytes(msg)

    def no_send_tag(self):
        """Skip OTR whitespace tagging to bots and services.

        Any nicks matching the otr.general.no_send_tag_regex config setting
        will not be tagged.
        """
        no_send_tag_regex = config_string('general.no_send_tag_regex')
        debug(('no_send_tag', no_send_tag_regex, self.peer_nick))
        if no_send_tag_regex:
            return re.match(no_send_tag_regex, self.peer_nick, re.IGNORECASE)

    def __repr__(self):
        return PYVER.to_str((
            '<{0} {1:x} peer_nick={c.peer_nick} peer_server={c.peer_server}>'
            ).format(self.__class__.__name__, id(self), c=self))

class IrcOtrAccount(potr.context.Account):
    """Account class for OTR over IRC."""

    contextclass = IrcContext

    PROTOCOL = 'irc'
    MAX_MSG_SIZE = 415

    def __init__(self, name):
        super(IrcOtrAccount, self).__init__(
            name, IrcOtrAccount.PROTOCOL, IrcOtrAccount.MAX_MSG_SIZE)

        self.nick, self.server = self.name.split('@', 1)

        # IRC messages cannot have newlines, OTR query and "no plugin" text
        # need to be one message
        self.defaultQuery = self.defaultQuery.replace("\n", ' ')

        self.key_file_path = private_key_file_path(name)
        self.fpr_file_path = os.path.join(OTR_DIR, '{0}.fpr'.format(name))

        self.load_trusts()

    def load_trusts(self):
        """Load trust data from the fingerprint file."""
        if os.path.exists(self.fpr_file_path):
            with open(self.fpr_file_path) as fpr_file:
                for line in fpr_file:
                    debug(('load trust check', line))

                    context, account, protocol, fpr, trust = \
                        PYVER.to_unicode(line[:-1]).split('\t')

                    if account == self.name and \
                            protocol == IrcOtrAccount.PROTOCOL:
                        debug(('set trust', context, fpr, trust))
                        self.setTrust(context, fpr, trust)

    def loadPrivkey(self):
        """Load key file.

        If no key file exists, load the default key. If there is no default
        key, a new key will be generated automatically by potr."""
        debug(('load private key', self.key_file_path))

        if os.path.exists(self.key_file_path):
            return read_private_key(self.key_file_path)
        else:
            default_key = config_string('general.defaultkey')
            if default_key:
                default_key_path = private_key_file_path(default_key)

                if os.path.exists(default_key_path):
                    shutil.copyfile(default_key_path, self.key_file_path)
                    return read_private_key(self.key_file_path)

    def savePrivkey(self):
        """Save key file."""
        debug(('save private key', self.key_file_path))

        with open(self.key_file_path, 'wb') as key_file:
            key_file.write(self.getPrivkey().serializePrivateKey())

    def saveTrusts(self):
        """Save trusts."""
        with open(self.fpr_file_path, 'w') as fpr_file:
            for uid, trusts in self.trusts.items():
                for fpr, trust in trusts.items():
                    debug(('trust write', uid, self.name,
                           IrcOtrAccount.PROTOCOL, fpr, trust))
                    fpr_file.write(PYVER.to_str('\t'.join(
                        (uid, self.name, IrcOtrAccount.PROTOCOL, fpr, trust))))
                    fpr_file.write('\n')

    def end_all_private(self):
        """End all currently encrypted conversations."""
        for context in self.ctxs.values():
            if context.is_encrypted():
                context.disconnect()

    def policy_config_option(self, policy):
        """Get the option name of a policy option for this account."""
        return config_prefix('.'.join([
            'policy', self.server, self.nick, policy.lower()]))

class IrcHTMLParser(PYVER.html_parser.HTMLParser):
    """A simple HTML parser that throws away anything but newlines and links"""

    @staticmethod
    def parse(data):
        """Create a temporary IrcHTMLParser and parse a single string"""
        parser = IrcHTMLParser(**PYVER.html_parser_init_kwargs)
        parser.feed(data)
        parser.close()
        return parser.result

    def reset(self):
        """Forget all state, called from __init__"""
        PYVER.html_parser.HTMLParser.reset(self)
        self.result = ''
        self.linktarget = ''
        self.linkstart = 0

    def handle_starttag(self, tag, attrs):
        """Called when a start tag is encountered"""
        if tag == 'br':
            self.result += '\n'
        elif tag == 'a':
            attrs = dict(attrs)
            if 'href' in attrs:
                self.result += '['
                self.linktarget = attrs['href']
                self.linkstart = len(self.result)

    def handle_endtag(self, tag):
        """Called when an end tag is encountered"""
        if tag == 'a':
            if self.linktarget:
                if self.result[self.linkstart:] == self.linktarget:
                    self.result += ']'
                else:
                    self.result += ']({0})'.format(self.linktarget)
                self.linktarget = ''

    def handle_data(self, data):
        """Called for character data (i.e. text)"""
        self.result += data

    def handle_entityref(self, name):
        """Called for entity references, such as &amp;"""
        try:
            self.result += PYVER.unichr(
                PYVER.html_entities.name2codepoint[name])
        except KeyError:
            self.result += '&{0};'.format(name)

    def handle_charref(self, name):
        """Called for character references, such as &#39;"""
        try:
            if name.startswith('x'):
                self.result += PYVER.unichr(int(name[1:], 16))
            else:
                self.result += PYVER.unichr(int(name))
        except ValueError:
            self.result += '&#{0};'.format(name)

class TableFormatter(object):
    """Format lists of string into aligned tables."""

    def __init__(self):
        self.rows = []
        self.max_widths = None

    def add_row(self, row):
        """Add a row to the table."""
        self.rows.append(row)
        row_widths = [len(s) for s in row]
        if self.max_widths is None:
            self.max_widths = row_widths
        else:
            self.max_widths = list(map(max, self.max_widths, row_widths))

    def format(self):
        """Return the formatted table as a string."""
        return '\n'.join([self.format_row(row) for row in self.rows])

    def format_row(self, row):
        """Format a single row as a string."""
        return ' |'.join(
            [s.ljust(self.max_widths[i]) for i, s in enumerate(row)])

def message_in_cb(data, modifier, modifier_data, string):
    """Incoming message callback"""
    debug(('message_in_cb', data, modifier, modifier_data, string))

    parsed = parse_irc_privmsg(
        PYVER.to_unicode(string), PYVER.to_unicode(modifier_data))
    debug(('parsed message', parsed))

    # skip processing messages to public channels
    if parsed['to_channel']:
        return string

    server = PYVER.to_unicode(modifier_data)

    context = get_server_context(server, parsed['from_nick'])

    context.in_assembler.add(parsed['text'])

    result = ''

    if context.in_assembler.is_done():
        try:
            msg, tlvs = context.receiveMessage(
                # potr expects bytes
                to_bytes(context.in_assembler.get()))

            debug(('receive', msg, tlvs))

            if msg:
                result = PYVER.to_str(build_privmsgs_in(
                    parsed['from'], parsed['to'],
                    context.msg_convert_in(msg)))

            context.handle_tlvs(tlvs)
        except potr.context.ErrorReceived as err:
            context.print_buffer('Received OTR error: {0}'.format(
                PYVER.to_unicode(err.args[0])), 'error')
        except potr.context.NotEncryptedError:
            context.print_buffer(
                'Received encrypted data but no private session established.',
                'warning')
        except potr.context.NotOTRMessage:
            result = string
        except potr.context.UnencryptedMessage as err:
            result = PYVER.to_str(build_privmsgs_in(
                parsed['from'], parsed['to'], PYVER.to_unicode(
                    msg_plain_from_irc(err.args[0])),
                'Unencrypted message received: '))

    weechat.bar_item_update(SCRIPT_NAME)

    return result

def message_out_cb(data, modifier, modifier_data, string):
    """Outgoing message callback."""
    result = ''

    # If any exception is raised in this function, WeeChat will not send the
    # outgoing message, which could be something that the user intended to be
    # encrypted. This paranoid exception handling ensures that the system
    # fails closed and not open.
    try:
        debug(('message_out_cb', data, modifier, modifier_data, string))

        parsed = parse_irc_privmsg(
            PYVER.to_unicode(string), PYVER.to_unicode(modifier_data))
        debug(('parsed message', parsed))

        # skip processing messages to public channels
        if parsed['to_channel']:
            return string

        server = PYVER.to_unicode(modifier_data)

        context = get_server_context(server, parsed['to_nick'])
        is_query = OTR_QUERY_RE.search(parsed['text'])

        parsed_text_bytes = to_bytes(parsed['text'])

        is_otr_message = \
            parsed_text_bytes[:len(potr.proto.OTRTAG)] == potr.proto.OTRTAG

        if is_otr_message and not is_query:
            if not has_otr_end(parsed['text']):
                debug('in OTR message')
                context.in_otr_message = True
            else:
                debug('complete OTR message')
            result = string
        elif context.in_otr_message:
            if has_otr_end(parsed['text']):
                context.in_otr_message = False
                debug('in OTR message end')
            result = string
        else:
            debug(('context send message', parsed['text'], parsed['to_nick'],
                   server))

            if context.policyOtrEnabled() and \
                not context.is_encrypted() and \
                not is_query and \
                context.getPolicy('require_encryption'):
                context.print_buffer(
                    'Your message will not be sent, because policy requires an '
                    'encrypted connection.', 'error')
                context.hint(
                    'Wait for the OTR connection or change the policy to allow '
                    'clear-text messages:\n'
                    '/otr policy require_encryption off')

            try:
                ret = context.sendMessage(
                    potr.context.FRAGMENT_SEND_ALL,
                    context.msg_convert_out(parsed['text']))

                if ret:
                    debug(('sendMessage returned', ret))
                    result = PYVER.to_str(
                        build_privmsg_out(
                            parsed['to_nick'], PYVER.to_unicode(ret)
                            ))

            except potr.context.NotEncryptedError as err:
                if err.args[0] == potr.context.EXC_FINISHED:
                    context.print_buffer(
                        'Your message was not sent. End your private '
                        'conversation:\n/otr finish',
                        'error')
                else:
                    raise

        weechat.bar_item_update(SCRIPT_NAME)
    # pylint: disable=bare-except
    except:
        try:
            print_buffer('', traceback.format_exc(), 'error')
            print_buffer('', 'Versions: {versions}'.format(
                versions=dependency_versions()), 'error')
            context.print_buffer(
                'Failed to send message. See core buffer for traceback.',
                'error')
        # pylint: disable=bare-except
        except:
            pass

    return result

def shutdown():
    """Script unload callback."""
    debug('shutdown')

    weechat.config_write(CONFIG_FILE)

    for account in ACCOUNTS.values():
        account.end_all_private()

    free_all_config()

    weechat.bar_item_remove(OTR_STATUSBAR)

    return weechat.WEECHAT_RC_OK

def command_cb(data, buf, args):
    """Parse and dispatch WeeChat OTR commands."""
    result = weechat.WEECHAT_RC_ERROR

    arg_parts = [PYVER.to_unicode(arg) for arg in shlex.split(args)]

    if len(arg_parts) in (1, 3) and arg_parts[0] in ('start', 'refresh'):
        nick, server = default_peer_args(arg_parts[1:3], buf)

        if nick is not None and server is not None:
            context = get_server_context(server, nick)
            # We need to wall disable_logging() here so that no OTR-related
            # buffer messages get logged at any point. disable_logging() will
            # be called again when effectively switching to encrypted, but
            # the previous_log_level we set here will be preserved for later
            # restoring.
            if not context.getPolicy('log'):
                context.previous_log_level = context.disable_logging()
            else:
                context.previous_log_level = context.get_log_level()

            context.hint(
                'Sending OTR query... Please await confirmation of the OTR '
                'session being started before sending a message.')
            if not context.getPolicy('send_tag'):
                context.hint(
                    'To try OTR on all conversations with {peer}: /otr '
                    'policy send_tag on'.format(peer=context.peer))

            privmsg(server, nick, '?OTR?')

            result = weechat.WEECHAT_RC_OK
    elif len(arg_parts) in (1, 3) and arg_parts[0] in ('finish', 'end'):
        nick, server = default_peer_args(arg_parts[1:3], buf)

        if nick is not None and server is not None:
            context = get_server_context(server, nick)
            context.disconnect()

            result = weechat.WEECHAT_RC_OK

    elif len(arg_parts) in (1, 3) and arg_parts[0] == 'status':
        nick, server = default_peer_args(arg_parts[1:3], buf)

        if nick is not None and server is not None:
            context = get_server_context(server, nick)
            if context.is_encrypted():
                context.print_buffer(
                    'This conversation is encrypted.', 'success')
                context.print_buffer("Your fingerprint is: {0}".format(
                    context.user.getPrivkey()))
                context.print_buffer("Your peer's fingerprint is: {0}".format(
                    potr.human_hash(context.crypto.theirPubkey.cfingerprint())))
                if context.is_verified():
                    context.print_buffer(
                        "The peer's identity has been verified.",
                        'success')
                else:
                    context.print_buffer(
                        "You have not verified the peer's identity yet.",
                        'warning')
            else:
                context.print_buffer(
                    "This current conversation is not encrypted.",
                    'warning')

            result = weechat.WEECHAT_RC_OK

    elif len(arg_parts) in range(2, 7) and arg_parts[0] == 'smp':
        action = arg_parts[1]

        if action == 'respond':
            # Check if nickname and server are specified
            if len(arg_parts) == 3:
                nick, server = default_peer_args([], buf)
                secret = arg_parts[2]
            elif len(arg_parts) == 5:
                nick, server = default_peer_args(arg_parts[2:4], buf)
                secret = arg_parts[4]
            else:
                return weechat.WEECHAT_RC_ERROR

            if secret:
                secret = PYVER.to_str(secret)

            context = get_server_context(server, nick)
            context.smpGotSecret(secret)

            result = weechat.WEECHAT_RC_OK

        elif action == 'ask':
            question = None
            secret = None

            # Nickname and server are not specified
            # Check whether it's a simple challenge or a question/answer request
            if len(arg_parts) == 3:
                nick, server = default_peer_args([], buf)
                secret = arg_parts[2]
            elif len(arg_parts) == 4:
                nick, server = default_peer_args([], buf)
                secret = arg_parts[3]
                question = arg_parts[2]

            # Nickname and server are specified
            # Check whether it's a simple challenge or a question/answer request
            elif len(arg_parts) == 5:
                nick, server = default_peer_args(arg_parts[2:4], buf)
                secret = arg_parts[4]
            elif len(arg_parts) == 6:
                nick, server = default_peer_args(arg_parts[2:4], buf)
                secret = arg_parts[5]
                question = arg_parts[4]
            else:
                return weechat.WEECHAT_RC_ERROR

            context = get_server_context(server, nick)

            if secret:
                secret = PYVER.to_str(secret)
            if question:
                question = PYVER.to_str(question)

            try:
                context.smpInit(secret, question)
            except potr.context.NotEncryptedError:
                context.print_buffer(
                    'There is currently no encrypted session with {0}.'.format(
                        context.peer), 'error')
            else:
                if question:
                    context.print_buffer('SMP challenge sent...')
                else:
                    context.print_buffer('SMP question sent...')
                context.in_smp = True
                result = weechat.WEECHAT_RC_OK

        elif action == 'abort':
            # Nickname and server are not specified
            if len(arg_parts) == 2:
                nick, server = default_peer_args([], buf)
            # Nickname and server are specified
            elif len(arg_parts) == 4:
                nick, server = default_peer_args(arg_parts[2:4], buf)
            else:
                return weechat.WEECHAT_RC_ERROR

            context = get_server_context(server, nick)

            if context.in_smp:
                try:
                    context.smpAbort()
                except potr.context.NotEncryptedError:
                    context.print_buffer(
                        'There is currently no encrypted session with {0}.'
                        .format(context.peer), 'error')
                else:
                    debug('SMP aborted')
                    context.smp_finish('SMP aborted.')
                    result = weechat.WEECHAT_RC_OK

    elif len(arg_parts) in (1, 3) and arg_parts[0] == 'trust':
        nick, server = default_peer_args(arg_parts[1:3], buf)

        if nick is not None and server is not None:
            context = get_server_context(server, nick)

            if context.crypto.theirPubkey is not None:
                context.setCurrentTrust('verified')
                context.print_buffer('{peer} is now authenticated.'.format(
                    peer=context.peer))

                weechat.bar_item_update(SCRIPT_NAME)
            else:
                context.print_buffer(
                    'No fingerprint for {peer}. Start an OTR conversation '
                    'first: /otr start'.format(peer=context.peer),
                    'error')

            result = weechat.WEECHAT_RC_OK
    elif len(arg_parts) in (1, 3) and arg_parts[0] == 'distrust':
        nick, server = default_peer_args(arg_parts[1:3], buf)

        if nick is not None and server is not None:
            context = get_server_context(server, nick)

            if context.crypto.theirPubkey is not None:
                context.setCurrentTrust('')
                context.print_buffer(
                    '{peer} is now de-authenticated.'.format(
                        peer=context.peer))

                weechat.bar_item_update(SCRIPT_NAME)
            else:
                context.print_buffer(
                    'No fingerprint for {peer}. Start an OTR conversation '
                    'first: /otr start'.format(peer=context.peer), 'error')

            result = weechat.WEECHAT_RC_OK

    elif len(arg_parts) in (1, 2) and arg_parts[0] == 'log':
        nick, server = default_peer_args([], buf)
        if len(arg_parts) == 1:
            if nick is not None and server is not None:
                context = get_server_context(server, nick)

                if context.is_encrypted():
                    if context.is_logged():
                        context.print_buffer(
                            'This conversation is currently being logged.',
                            'warning')
                        result = weechat.WEECHAT_RC_OK

                    else:
                        context.print_buffer(
                            'This conversation is currently NOT being logged.')
                        result = weechat.WEECHAT_RC_OK
                else:
                    context.print_buffer(
                        'OTR LOG: Not in an OTR session', 'error')
                    result = weechat.WEECHAT_RC_OK

            else:
                print_buffer('', 'OTR LOG: Not in an OTR session', 'error')
                result = weechat.WEECHAT_RC_OK

        if len(arg_parts) == 2:
            if nick is not None and server is not None:
                context = get_server_context(server, nick)

                if arg_parts[1] == 'start' and \
                        not context.is_logged() and \
                        context.is_encrypted():
                    if context.previous_log_level is None:
                        context.previous_log_level = context.get_log_level()
                    context.print_buffer(
                        'From this point on, this conversation will be '
                        'logged. Please keep in mind that by doing so you '
                        'are potentially putting yourself and your '
                        'interlocutor at risk. You can disable this by doing '
                        '/otr log stop',
                        'warning')
                    weechat.command(buf, '/mute logger set 9')
                    result = weechat.WEECHAT_RC_OK

                elif arg_parts[1] == 'stop' and \
                        context.is_logged() and \
                        context.is_encrypted():
                    if context.previous_log_level is None:
                        context.previous_log_level = context.get_log_level()
                    weechat.command(buf, '/mute logger set 0')
                    context.print_buffer(
                        'From this point on, this conversation will NOT be '
                        'logged ANYMORE.')
                    result = weechat.WEECHAT_RC_OK

                elif not context.is_encrypted():
                    context.print_buffer(
                        'OTR LOG: Not in an OTR session', 'error')
                    result = weechat.WEECHAT_RC_OK

                else:
                    # Don't need to do anything.
                    result = weechat.WEECHAT_RC_OK

            else:
                print_buffer('', 'OTR LOG: Not in an OTR session', 'error')

    elif len(arg_parts) in (1, 2, 3, 4) and arg_parts[0] == 'policy':
        if len(arg_parts) == 1:
            nick, server = default_peer_args([], buf)

            if nick is not None and server is not None:
                context = get_server_context(server, nick)

                context.print_buffer(context.format_policies())
            else:
                prnt('', format_default_policies())

            result = weechat.WEECHAT_RC_OK

        elif len(arg_parts) == 2 and arg_parts[1].lower() == 'default':
            nick, server = default_peer_args([], buf)

            if nick is not None and server is not None:
                context = get_server_context(server, nick)

                context.print_buffer(format_default_policies())
            else:
                prnt('', format_default_policies())

            result = weechat.WEECHAT_RC_OK

        elif len(arg_parts) == 3 and arg_parts[1].lower() in POLICIES:
            nick, server = default_peer_args([], buf)

            if nick is not None and server is not None:
                context = get_server_context(server, nick)

                policy_var = context.policy_config_option(arg_parts[1].lower())

                command('', '/set {policy} {value}'.format(
                    policy=policy_var,
                    value=arg_parts[2]))

                context.print_buffer(context.format_policies())

                result = weechat.WEECHAT_RC_OK

        elif len(arg_parts) == 4 and \
            arg_parts[1].lower() == 'default' and \
            arg_parts[2].lower() in POLICIES:
            nick, server = default_peer_args([], buf)

            policy_var = "otr.policy.default." + arg_parts[2].lower()

            command('', '/set {policy} {value}'.format(
                policy=policy_var,
                value=arg_parts[3]))

            if nick is not None and server is not None:
                context = get_server_context(server, nick)

                context.print_buffer(format_default_policies())
            else:
                prnt('', format_default_policies())

            result = weechat.WEECHAT_RC_OK
    elif len(arg_parts) in (1, 2) and arg_parts[0] == 'fingerprint':
        if len(arg_parts) == 1:
            show_account_fingerprints()
            result = weechat.WEECHAT_RC_OK
        elif len(arg_parts) == 2:
            if arg_parts[1] == 'all':
                show_peer_fingerprints()
            else:
                show_peer_fingerprints(grep=arg_parts[1])
            result = weechat.WEECHAT_RC_OK

    return result

def otr_statusbar_cb(data, item, window):
    """Update the statusbar."""
    if window:
        buf = weechat.window_get_pointer(window, 'buffer')
    else:
        # If the bar item is in a root bar that is not in a window, window
        # will be empty.
        buf = weechat.current_buffer()

    if not buffer_is_private(buf):
        return ''

    local_user = irc_user(
        buffer_get_string(buf, 'localvar_nick'),
        buffer_get_string(buf, 'localvar_server'))

    remote_user = irc_user(
        buffer_get_string(buf, 'localvar_channel'),
        buffer_get_string(buf, 'localvar_server'))

    context = get_context(local_user, remote_user)

    encrypted_str = config_string('look.bar.state.encrypted')
    unencrypted_str = config_string('look.bar.state.unencrypted')
    authenticated_str = config_string('look.bar.state.authenticated')
    unauthenticated_str = config_string('look.bar.state.unauthenticated')
    logged_str = config_string('look.bar.state.logged')
    notlogged_str = config_string('look.bar.state.notlogged')

    bar_parts = []

    if context.is_encrypted():
        if encrypted_str:
            bar_parts.append(''.join([
                config_color('status.encrypted'),
                encrypted_str,
                config_color('status.default')]))

        if context.is_verified():
            if authenticated_str:
                bar_parts.append(''.join([
                    config_color('status.authenticated'),
                    authenticated_str,
                    config_color('status.default')]))
        elif unauthenticated_str:
            bar_parts.append(''.join([
                config_color('status.unauthenticated'),
                unauthenticated_str,
                config_color('status.default')]))

        if context.is_logged():
            if logged_str:
                bar_parts.append(''.join([
                    config_color('status.logged'),
                    logged_str,
                    config_color('status.default')]))
        elif notlogged_str:
            bar_parts.append(''.join([
                config_color('status.notlogged'),
                notlogged_str,
                config_color('status.default')]))

    elif unencrypted_str:
        bar_parts.append(''.join([
            config_color('status.unencrypted'),
            unencrypted_str,
            config_color('status.default')]))

    result = config_string('look.bar.state.separator').join(bar_parts)

    if result:
        result = '{color}{prefix}{result}'.format(
            color=config_color('status.default'),
            prefix=config_string('look.bar.prefix'),
            result=result)

    if context.is_encrypted():
        weechat.buffer_set(buf, 'localvar_set_otr_encrypted', 'true')
    else:
        weechat.buffer_set(buf, 'localvar_set_otr_encrypted', 'false')

    if context.is_verified():
        weechat.buffer_set(buf, 'localvar_set_otr_authenticated', 'true')
    else:
        weechat.buffer_set(buf, 'localvar_set_otr_authenticated', 'false')

    if context.is_logged():
        weechat.buffer_set(buf, 'localvar_set_otr_logged', 'true')
    else:
        weechat.buffer_set(buf, 'localvar_set_otr_logged', 'false')

    return result

def bar_config_update_cb(data, option):
    """Callback for updating the status bar when its config changes."""
    weechat.bar_item_update(SCRIPT_NAME)

    return weechat.WEECHAT_RC_OK

def policy_completion_cb(data, completion_item, buf, completion):
    """Callback for policy tab completion."""
    for policy in POLICIES:
        weechat.hook_completion_list_add(
            completion, policy, 0, weechat.WEECHAT_LIST_POS_SORT)

    return weechat.WEECHAT_RC_OK

def policy_create_option_cb(data, config_file, section, name, value):
    """Callback for creating a new policy option when the user sets one
    that doesn't exist."""
    weechat.config_new_option(
        config_file, section, name, 'boolean', '', '', 0, 0, value, value, 0,
        '', '', '', '', '', '')

    return weechat.WEECHAT_CONFIG_OPTION_SET_OK_CHANGED

def logger_level_update_cb(data, option, value):
    """Callback called when any logger level changes."""
    weechat.bar_item_update(SCRIPT_NAME)

    return weechat.WEECHAT_RC_OK

def buffer_switch_cb(data, signal, signal_data):
    """Callback for buffer switched.

    Used for updating the status bar item when it is in a root bar.
    """
    weechat.bar_item_update(SCRIPT_NAME)

    return weechat.WEECHAT_RC_OK

def buffer_closing_cb(data, signal, signal_data):
    """Callback for buffer closed.

    It closes the OTR session when the buffer is about to be closed.
    """
    result = weechat.WEECHAT_RC_ERROR
    nick, server = default_peer_args([], signal_data)

    if nick is not None and server is not None:
        context = get_server_context(server, nick)
        context.disconnect()

        result = weechat.WEECHAT_RC_OK
    return result

def init_config():
    """Set up configuration options and load config file."""
    global CONFIG_FILE
    CONFIG_FILE = weechat.config_new(SCRIPT_NAME, '', '')

    global CONFIG_SECTIONS
    CONFIG_SECTIONS = {}

    CONFIG_SECTIONS['general'] = weechat.config_new_section(
        CONFIG_FILE, 'general', 0, 0, '', '', '', '', '', '', '', '', '', '')

    for option, typ, desc, default in [
            ('debug',
             'boolean',
             'OTR script debugging',
             'off'),
            ('hints',
             'boolean',
             'Give helpful hints how to use this script and how to stay '
             'secure while using OTR (recommended)',
             'on'),
            ('defaultkey',
             'string',
             'default private key to use for new accounts (nick@server)',
             ''),
            ('no_send_tag_regex',
             'string',
             'do not OTR whitespace tag messages to nicks matching this regex '
             '(case insensitive)',
             '^(alis|chanfix|global|.+serv|\*.+)$'),
        ]:
        weechat.config_new_option(
            CONFIG_FILE, CONFIG_SECTIONS['general'], option, typ, desc, '', 0,
            0, default, default, 0, '', '', '', '', '', '')

    CONFIG_SECTIONS['color'] = weechat.config_new_section(
        CONFIG_FILE, 'color', 0, 0, '', '', '', '', '', '', '', '', '', '')

    for option, desc, default, update_cb in [
            ('status.default',
             'status bar default color',
             'default',
             'bar_config_update_cb'),
            ('status.encrypted',
             'status bar encrypted indicator color',
             'green',
             'bar_config_update_cb'),
            ('status.unencrypted',
             'status bar unencrypted indicator color',
             'lightred',
             'bar_config_update_cb'),
            ('status.authenticated',
             'status bar authenticated indicator color',
             'green',
             'bar_config_update_cb'),
            ('status.unauthenticated',
             'status bar unauthenticated indicator color',
             'lightred',
             'bar_config_update_cb'),
            ('status.logged',
             'status bar logged indicator color',
             'lightred',
             'bar_config_update_cb'),
            ('status.notlogged',
             'status bar not logged indicator color',
             'green',
             'bar_config_update_cb'),
            ('buffer.hint',
             'text color for hints',
             'lightblue',
             ''),
            ('buffer.info',
             'text color for informational messages',
             'default',
             ''),
            ('buffer.success',
             'text color for success messages',
             'lightgreen',
             ''),
            ('buffer.warning',
             'text color for warnings',
             'yellow',
             ''),
            ('buffer.error',
             'text color for errors',
             'lightred',
             ''),
        ]:
        weechat.config_new_option(
            CONFIG_FILE, CONFIG_SECTIONS['color'], option, 'color', desc, '', 0,
            0, default, default, 0, '', '', update_cb, '', '', '')

    CONFIG_SECTIONS['look'] = weechat.config_new_section(
        CONFIG_FILE, 'look', 0, 0, '', '', '', '', '', '', '', '', '', '')

    for option, desc, default, update_cb in [
            ('bar.prefix',
             'prefix for OTR status bar item',
             'OTR:',
             'bar_config_update_cb'),
            ('bar.state.encrypted',
             'shown in status bar when conversation is encrypted',
             'SEC',
             'bar_config_update_cb'),
            ('bar.state.unencrypted',
             'shown in status bar when conversation is not encrypted',
             '!SEC',
             'bar_config_update_cb'),
            ('bar.state.authenticated',
             'shown in status bar when peer is authenticated',
             'AUTH',
             'bar_config_update_cb'),
            ('bar.state.unauthenticated',
             'shown in status bar when peer is not authenticated',
             '!AUTH',
             'bar_config_update_cb'),
            ('bar.state.logged',
             'shown in status bar when peer conversation is being logged to '
             'disk',
             'LOG',
             'bar_config_update_cb'),
            ('bar.state.notlogged',
             'shown in status bar when peer conversation is not being logged '
             'to disk',
             '!LOG',
             'bar_config_update_cb'),
            ('bar.state.separator',
             'separator for states in the status bar',
             ',',
             'bar_config_update_cb'),
            ('prefix',
             'prefix used for messages from otr (note: content is evaluated, '
             'see /help eval)',
             '${color:default}:! ${color:brown}otr${color:default} !:',
             ''),
        ]:
        weechat.config_new_option(
            CONFIG_FILE, CONFIG_SECTIONS['look'], option, 'string', desc, '',
            0, 0, default, default, 0, '', '', update_cb, '', '', '')

    CONFIG_SECTIONS['policy'] = weechat.config_new_section(
        CONFIG_FILE, 'policy', 1, 1, '', '', '', '', '', '',
        'policy_create_option_cb', '', '', '')

    for option, desc, default in [
            ('default.allow_v2',
             'default allow OTR v2 policy',
             'on'),
            ('default.require_encryption',
             'default require encryption policy',
             'off'),
            ('default.log',
             'default enable logging to disk',
             'off'),
            ('default.send_tag',
             'default send tag policy',
             'off'),
            ('default.html_escape',
             'default HTML escape policy',
             'off'),
            ('default.html_filter',
             'default HTML filter policy',
             'on'),
        ]:
        weechat.config_new_option(
            CONFIG_FILE, CONFIG_SECTIONS['policy'], option, 'boolean', desc, '',
            0, 0, default, default, 0, '', '', '', '', '', '')

    weechat.config_read(CONFIG_FILE)

def free_all_config():
    """Free all config options, sections and config file."""
    for section in CONFIG_SECTIONS.values():
        weechat.config_section_free_options(section)
        weechat.config_section_free(section)

    weechat.config_free(CONFIG_FILE)

def create_dir():
    """Create the OTR subdirectory in the WeeChat config directory if it does
    not exist."""
    if not os.path.exists(OTR_DIR):
        weechat.mkdir_home(OTR_DIR_NAME, 0o700)

def git_info():
    """If this script is part of a git repository return the repo state."""
    result = None
    script_dir = os.path.dirname(os.path.realpath(__file__))
    git_dir = os.path.join(script_dir, '.git')
    if os.path.isdir(git_dir):
        import subprocess
        try:
            # We can't use check_output here without breaking compatibility
            # for Python 2.6, but we ignore the return value anyway, so Popen
            # is only slightly more complicated:
            process = subprocess.Popen([
                'git',
                '--git-dir', git_dir,
                '--work-tree', script_dir,
                'describe', '--dirty', '--always',
                ], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output = process.communicate()[0]
            if output:
                result = PYVER.to_unicode(output).lstrip('v').rstrip()
        except OSError:
            pass

    return result

def weechat_version_ok():
    """Check if the WeeChat version is compatible with this script.

    If WeeChat version < 0.4.2 log an error to the core buffer and return
    False. Otherwise return True.
    """
    weechat_version = weechat.info_get('version_number', '') or 0
    if int(weechat_version) < 0x00040200:
        error_message = (
            '{script_name} requires WeeChat version >= 0.4.2. The current '
            'version is {current_version}.').format(
                script_name=SCRIPT_NAME,
                current_version=weechat.info_get('version', ''))
        prnt('', error_message)
        return False
    return True

SCRIPT_VERSION = git_info() or SCRIPT_VERSION

def dependency_versions():
    """Return a string containing the versions of all dependencies."""
    return ('weechat-otr {script_version}, '
            'potr {potr_major}.{potr_minor}.{potr_patch}-{potr_sub}, '
            'Python {python_version}, '
            'WeeChat {weechat_version}'
           ).format(
               script_version=SCRIPT_VERSION,
               potr_major=potr.VERSION[0],
               potr_minor=potr.VERSION[1],
               potr_patch=potr.VERSION[2],
               potr_sub=potr.VERSION[3],
               python_version=platform.python_version(),
               weechat_version=weechat.info_get('version', ''))

def excepthook(typ, value, tracebak):
    """Add dependency versions to tracebacks."""
    sys.stderr.write('Versions: ')
    sys.stderr.write(dependency_versions())
    sys.stderr.write('\n')

    sys.__excepthook__(typ, value, tracebak)

sys.excepthook = excepthook

if weechat.register(
        SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENCE, SCRIPT_DESC,
        'shutdown', ''):
    if weechat_version_ok():
        init_config()

        OTR_DIR = os.path.join(info_get('weechat_dir', ''), OTR_DIR_NAME)
        create_dir()

        ACCOUNTS = AccountDict()

        weechat.hook_modifier('irc_in_privmsg', 'message_in_cb', '')
        weechat.hook_modifier('irc_out_privmsg', 'message_out_cb', '')

        weechat.hook_command(
            SCRIPT_NAME, SCRIPT_HELP,
            'start [NICK SERVER] || '
            'refresh [NICK SERVER] || '
            'finish [NICK SERVER] || '
            'end [NICK SERVER] || '
            'status [NICK SERVER] || '
            'smp ask [NICK SERVER] [QUESTION] SECRET || '
            'smp respond [NICK SERVER] SECRET || '
            'smp abort [NICK SERVER] || '
            'trust [NICK SERVER] || '
            'distrust [NICK SERVER] || '
            'log [start|stop] || '
            'policy [POLICY on|off] || '
            'fingerprint [SEARCH|all]',
            '',
            'start %(nick) %(irc_servers) %-||'
            'refresh %(nick) %(irc_servers) %-||'
            'finish %(nick) %(irc_servers) %-||'
            'end %(nick) %(irc_servers) %-||'
            'status %(nick) %(irc_servers) %-||'
            'smp ask|respond %(nick) %(irc_servers) %-||'
            'smp abort %(nick) %(irc_servers) %-||'
            'trust %(nick) %(irc_servers) %-||'
            'distrust %(nick) %(irc_servers) %-||'
            'log start|stop %-||'
            'policy %(otr_policy) on|off %-||'
            'fingerprint all %-||',
            'command_cb',
            '')

        weechat.hook_completion(
            'otr_policy', 'OTR policies', 'policy_completion_cb', '')

        weechat.hook_config('logger.level.irc.*', 'logger_level_update_cb', '')

        weechat.hook_signal('buffer_switch', 'buffer_switch_cb', '')
        weechat.hook_signal('buffer_closing', 'buffer_closing_cb', '')

        OTR_STATUSBAR = weechat.bar_item_new(
            SCRIPT_NAME, 'otr_statusbar_cb', '')
        weechat.bar_item_update(SCRIPT_NAME)



# -*- coding: utf-8 -*-
###
# Copyright (c) 2009-2011 by Elián Hanisch <lambdae2@gmail.com>
# Copyright (c) 2013 by Filip H.F. "FiXato" Slagter <fixato+weechat@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###

###
# Prints user's country and local time information in
# whois/whowas replies (for WeeChat 0.3.*)
#
#   This script uses MaxMind's GeoLite database from
#   http://www.maxmind.com/app/geolitecountry
#
#   This script depends in pytz third party module for retrieving
#   timezone information for a given country. Without it the local time
#   for a user won't be displayed.
#   Get it from http://pytz.sourceforge.net or from your distro packages,
#   python-tz in Ubuntu/Debian
#
#   Commands:
#   * /country
#     Prints country for a given ip, uri or nick. See /help country
#
#   Settings:
#   * plugins.var.python.country.show_in_whois:
#     If 'off' /whois or /whowas replies won't contain country information.
#     Valid values: on, off
#   * plugins.var.python.country.show_localtime:
#     If 'off' timezone and local time infomation won't be looked for.
#     Valid values: on, off
#
#
#   TODO
#   * Add support for IPv6 addresses
#
#
#   History:
#   2013-04-28
#   version 0.6:
#   * Improved support for target msgbuffer. Takes the following settings into account:
#       - irc.msgbuffer.whois
#       - irc.msgbuffer.$servername.whois
#       - irc.look.msgbuffer_fallback
#
#   2011-08-14
#   version 0.5:
#   * make time format configurable.
#   * print to private buffer based on msgbuffer setting.
#
#   2011-01-09
#   version 0.4.1: bug fixes
#
#   2010-11-15
#   version 0.4:
#   * support for users using webchat (at least in freenode)
#   * enable Archlinux workaround.
#
#   2010-01-11
#   version 0.3.1: bug fix
#   * irc_nick infolist wasn't freed in get_host_by_nick()
#
#   2009-12-12
#   version 0.3: update WeeChat site.
#
#   2009-09-17
#   version 0.2: added timezone and local time information.
#
#   2009-08-24
#   version 0.1.1: fixed python 2.5 compatibility.
#
#   2009-08-21
#   version 0.1: initial release.
#
###

SCRIPT_NAME    = "country"
SCRIPT_AUTHOR  = "Elián Hanisch <lambdae2@gmail.com>"
SCRIPT_VERSION = "0.6"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC    = "Prints user's country and local time in whois replies"
SCRIPT_COMMAND = "country"

try:
    import weechat
    from weechat import WEECHAT_RC_OK, prnt
    import_ok = True
except ImportError:
    print "This script must be run under WeeChat."
    print "Get WeeChat now at: http://www.weechat.org/"
    import_ok = False

try:
    import pytz, datetime
    pytz_module = True
except:
    pytz_module = False

import os, re, socket

### ip database
database_url = 'http://geolite.maxmind.com/download/geoip/database/GeoIPCountryCSV.zip'
database_file = 'GeoIPCountryWhois.csv'

### config
settings = {
        'time_format': '%x %X %Z',
        'show_in_whois': 'on',
        'show_localtime': 'on'
        }

boolDict = {'on':True, 'off':False}
def get_config_boolean(config):
    value = weechat.config_get_plugin(config)
    try:
        return boolDict[value]
    except KeyError:
        default = settings[config]
        error("Error while fetching config '%s'. Using default value '%s'." %(config, default))
        error("'%s' is invalid, allowed: 'on', 'off'" %value)
        return boolDict[default]

### messages

script_nick = SCRIPT_NAME
def error(s, buffer=''):
    """Error msg"""
    prnt(buffer, '%s%s %s' % (weechat.prefix('error'), script_nick, s))
    if weechat.config_get_plugin('debug'):
        import traceback
        if traceback.sys.exc_type:
            trace = traceback.format_exc()
            prnt('', trace)

def say(s, buffer=''):
    """normal msg"""
    prnt(buffer, '%s\t%s' % (script_nick, s))

def whois(nick, string, buffer=''):
    """Message formatted like a whois reply."""
    prefix_network = weechat.prefix('network')
    color_delimiter = weechat.color('chat_delimiters')
    color_nick = weechat.color('chat_nick')
    prnt(buffer, '%s%s[%s%s%s] %s' % (prefix_network,
                                      color_delimiter,
                                      color_nick,
                                      nick,
                                      color_delimiter,
                                      string))

def string_country(country, code):
    """Format for country info string."""
    color_delimiter = weechat.color('chat_delimiters')
    color_chat = weechat.color('chat')
    return '%s%s %s(%s%s%s)' % (color_chat,
                                country,
                                color_delimiter,
                                color_chat,
                                code,
                                color_delimiter)

def string_time(dt):
    """Format for local time info string."""
    if not dt: return '--'
    color_delimiter = weechat.color('chat_delimiters')
    color_chat = weechat.color('chat')
    date = dt.strftime(weechat.config_get_plugin("time_format"))
    tz = dt.strftime('UTC%z')
    return '%s%s %s(%s%s%s)' % (color_chat,
                                date,
                                color_delimiter,
                                color_chat,
                                tz,
                                color_delimiter)

### functions
def get_script_dir():
    """Returns script's dir, creates it if needed."""
    script_dir = weechat.info_get('weechat_dir', '')
    script_dir = os.path.join(script_dir, 'country')
    if not os.path.isdir(script_dir):
        os.makedirs(script_dir)
    return script_dir

ip_database = ''
def check_database():
    """Check if there's a database already installed."""
    global ip_database
    if not ip_database:
        ip_database = os.path.join(get_script_dir(), database_file)
    return os.path.isfile(ip_database)

timeout = 1000*60*10
hook_download = ''
def update_database():
    """Downloads and uncompress the database."""
    global hook_download, ip_database
    if not ip_database:
        check_database()
    if hook_download:
        weechat.unhook(hook_download)
        hook_download = ''
    script_dir = get_script_dir()
    say("Downloading IP database...")
    python_bin = weechat.info_get('python2_bin', '') or 'python'
    hook_download = weechat.hook_process(
            python_bin + " -c \"\n"
            "import urllib2, zipfile, os, sys\n"
            "try:\n"
            "   temp = os.path.join('%(script_dir)s', 'temp.zip')\n"
            "   try:\n"
            "       zip = urllib2.urlopen('%(url)s', timeout=10)\n"
            "   except TypeError: # python2.5\n"
            "       import socket\n"
            "       socket.setdefaulttimeout(10)\n"
            "       zip = urllib2.urlopen('%(url)s')\n"
            "   fd = open(temp, 'w')\n"
            "   fd.write(zip.read())\n"
            "   fd.close()\n"
            "   print 'Download complete, uncompressing...'\n"
            "   zip = zipfile.ZipFile(temp)\n"
            "   try:\n"
            "       zip.extractall(path='%(script_dir)s')\n"
            "   except AttributeError: # python2.5\n"
            "       fd = open('%(ip_database)s', 'w')\n"
            "       fd.write(zip.read('%(database_file)s'))\n"
            "       fd.close()\n"
            "   os.remove(temp)\n"
            "except Exception, e:\n"
            "   print >>sys.stderr, e\n\"" % {'url':database_url,
                                              'script_dir':script_dir,
                                              'ip_database':ip_database,
                                              'database_file':database_file
                                              },
            timeout, 'update_database_cb', '')

process_stderr = ''
def update_database_cb(data, command, rc, stdout, stderr):
    """callback for our database download."""
    global hook_download, process_stderr
    #debug("%s @ stderr: '%s', stdout: '%s'" %(rc, stderr.strip('\n'), stdout.strip('\n')))
    if stdout:
        say(stdout)
    if stderr:
        process_stderr += stderr
    if int(rc) >= 0:
        if process_stderr:
            error(process_stderr)
            process_stderr = ''
        else:
            say('Success.')
        hook_download = ''
    return WEECHAT_RC_OK

hook_get_ip = ''
def get_ip_process(host):
    """Resolves host to ip."""
    # because getting the ip might take a while, we must hook a process so weechat doesn't hang.
    global hook_get_ip
    if hook_get_ip:
        weechat.unhook(hook_get_ip)
        hook_get_ip = ''
    python_bin = weechat.info_get('python2_bin', '') or 'python'
    hook_get_ip = weechat.hook_process(
            python_bin + " -c \"\n"
            "import socket, sys\n"
            "try:\n"
            "   ip = socket.gethostbyname('%(host)s')\n"
            "   print ip\n"
            "except Exception, e:\n"
            "   print >>sys.stderr, e\n\"" %{'host':host},
            timeout, 'get_ip_process_cb', '')

def get_ip_process_cb(data, command, rc, stdout, stderr):
    """Called when uri resolve finished."""
    global hook_get_ip, reply_wrapper
    #debug("%s @ stderr: '%s', stdout: '%s'" %(rc, stderr.strip('\n'), stdout.strip('\n')))
    if stdout and reply_wrapper:
        code, country = search_in_database(stdout[:-1])
        reply_wrapper(code, country)
        reply_wrapper = None
    if stderr and reply_wrapper:
        reply_wrapper(*unknown)
        reply_wrapper = None
    if int(rc) >= 0:
        hook_get_ip = ''
    return WEECHAT_RC_OK

def is_ip(s):
    """Returns whether or not a given string is an IPV4 address."""
    try:
        return bool(socket.inet_aton(s))
    except socket.error:
        return False

_valid_label = re.compile(r'^([\da-z]|[\da-z][-\da-z]*[\da-z])$', re.I)
def is_domain(s):
    """
    Checks if 's' is a valid domain."""
    if not s or len(s) > 255:
        return False
    labels = s.split('.')
    if len(labels) < 2:
        return False
    for label in labels:
        if not label or len(label) > 63 \
                or not _valid_label.match(label):
            return False
    return True

def hex_to_ip(s):
    """
    '7f000001' => '127.0.0.1'"""
    try:
        ip = map(lambda n: s[n:n+2], range(0, len(s), 2))
        ip = map(lambda n: int(n, 16), ip)
        return '.'.join(map(str, ip))
    except:
        return ''

def get_userhost_from_nick(buffer, nick):
    """Return host of a given nick in buffer."""
    channel = weechat.buffer_get_string(buffer, 'localvar_channel')
    server = weechat.buffer_get_string(buffer, 'localvar_server')
    if channel and server:
        infolist = weechat.infolist_get('irc_nick', '', '%s,%s' %(server, channel))
        if infolist:
            try:
                while weechat.infolist_next(infolist):
                    name = weechat.infolist_string(infolist, 'name')
                    if nick == name:
                        return weechat.infolist_string(infolist, 'host')
            finally:
                weechat.infolist_free(infolist)
    return ''

def get_ip_from_userhost(user, host):
    ip = get_ip_from_host(host)
    if ip:
        return ip
    ip = get_ip_from_user(user)
    if ip:
        return ip
    return host

def get_ip_from_host(host):
    if is_domain(host):
        return host
    else:
        if host.startswith('gateway/web/freenode/ip.'):
            ip = host.split('.', 1)[1]
            return ip

def get_ip_from_user(user):
    user = user[-8:] # only interested in the last 8 chars
    if len(user) == 8:
        ip = hex_to_ip(user)
        if ip and is_ip(ip):
            return ip

def sum_ip(ip):
    """Converts the ip number from dot-decimal notation to decimal."""
    L = map(int, ip.split('.'))
    return L[0]*16777216 + L[1]*65536 + L[2]*256 + L[3]

unknown = ('--', 'unknown')
def search_in_database(ip):
    """
    search_in_database(ip_number) => (code, country)
    returns ('--', 'unknown') if nothing found
    """
    import csv
    global ip_database
    if not ip or not ip_database:
        return unknown
    try:
        # do a binary search.
        n = sum_ip(ip)
        fd = open(ip_database)
        reader = csv.reader(fd)
        max = os.path.getsize(ip_database)
        last_high = last_low = min = 0
        while True:
            mid = (max + min)/2
            fd.seek(mid)
            fd.readline() # move cursor to next line
            _, _, low, high, code, country = reader.next()
            if low == last_low and high == last_high:
                break
            if n < long(low):
                max = mid
            elif n > long(high):
                min = mid
            elif n > long(low) and n < long(high):
                return (code, country)
            else:
                break
            last_low, last_high = low, high
    except StopIteration:
        pass
    return unknown

def print_country(host, buffer, quiet=False, broken=False, nick=''):
    """
    Prints country and local time for a given host, if quiet is True prints only if there's a match,
    if broken is True reply will be split in two messages.
    """
    #debug('host: ' + host)
    def reply_country(code, country):
        if quiet and code == '--':
            return
        if pytz_module and get_config_boolean('show_localtime') and code != '--':
            dt = get_country_datetime(code)
            if broken:
                whois(nick or host, string_country(country, code), buffer)
                whois(nick or host, string_time(dt), buffer)
            else:
                s = '%s - %s' %(string_country(country, code), string_time(dt))
                whois(nick or host, s, buffer)
        else:
            whois(nick or host, string_country(country, code), buffer)

    if is_ip(host):
        # good, got an ip
        code, country = search_in_database(host)
    elif is_domain(host):
        # try to resolve uri
        global reply_wrapper
        reply_wrapper = reply_country
        get_ip_process(host)
        return
    else:
        # probably a cloak or ipv6
        code, country = unknown
    reply_country(code, country)

### timezone
def get_country_datetime(code):
    """Get datetime object with country's timezone."""
    try:
        tzname = pytz.country_timezones(code)[0]
        tz = pytz.timezone(tzname)
        return datetime.datetime.now(tz)
    except:
        return None

### commands
def cmd_country(data, buffer, args):
    """Shows country and local time for a given ip, uri or nick."""
    if not args:
        weechat.command('', '/HELP %s' %SCRIPT_COMMAND)
        return WEECHAT_RC_OK
    if ' ' in args:
        # picks the first argument only
        args = args[:args.find(' ')]
    if args == 'update':
        update_database()
    else:
        if not check_database():
            error("IP database not found. You must download a database with '/country update' before "
                    "using this script.", buffer)
            return WEECHAT_RC_OK
        #check if is a nick
        userhost = get_userhost_from_nick(buffer, args)
        if userhost:
            host = get_ip_from_userhost(*userhost.split('@'))
        else:
            host = get_ip_from_userhost(args, args)
        print_country(host, buffer)
    return WEECHAT_RC_OK

def find_buffer(server, nick, message_type='whois'):
    # See if there is a target msgbuffer set for this server
    msgbuffer = weechat.config_string(weechat.config_get('irc.msgbuffer.%s.%s' % (server, message_type)))
    # No whois msgbuffer for this server; use the global setting
    if msgbuffer == '':
        msgbuffer = weechat.config_string(weechat.config_get('irc.msgbuffer.%s' % message_type))

    # Use the fallback msgbuffer setting if private buffer doesn't exist
    if msgbuffer == 'private':
        buffer = weechat.buffer_search('irc', '%s.%s' %(server, nick))
        if buffer != '':
            return buffer
        else:
            msgbuffer = weechat.config_string(weechat.config_get('irc.look.msgbuffer_fallback'))

    # Find the appropriate buffer
    if msgbuffer == "current":
        return weechat.current_buffer()
    elif msgbuffer == "weechat":
        return weechat.buffer_search_main()
    else:
        return weechat.buffer_search('irc', 'server.%s' % server)

### signal callbacks
def whois_cb(data, signal, signal_data):
    """function for /WHOIS"""
    if not get_config_boolean('show_in_whois') or not check_database():
        return WEECHAT_RC_OK
    nick, user, host = signal_data.split()[3:6]
    server = signal[:signal.find(',')]
    #debug('%s | %s | %s' %(data, signal, signal_data))
    host = get_ip_from_userhost(user, host)
    print_country(host, find_buffer(server, nick), quiet=True, broken=True, nick=nick)
    return WEECHAT_RC_OK

### main
if import_ok and weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE,
        SCRIPT_DESC, '', ''):

    # colors
    color_delimiter = weechat.color('chat_delimiters')
    color_chat_nick = weechat.color('chat_nick')
    color_reset     = weechat.color('reset')

    # pretty [SCRIPT_NAME]
    script_nick = '%s[%s%s%s]%s' % (color_delimiter,
                                    color_chat_nick,
                                    SCRIPT_NAME,
                                    color_delimiter,
                                    color_reset)

    weechat.hook_signal('*,irc_in2_311', 'whois_cb', '') # /whois
    weechat.hook_signal('*,irc_in2_314', 'whois_cb', '') # /whowas
    weechat.hook_command('country', cmd_country.__doc__, 'update | (nick|ip|uri)',
            "       update: Downloads/updates ip database with country codes.\n"
            "nick, ip, uri: Gets country and local time for a given ip, domain or nick.",
            'update||%(nick)', 'cmd_country', '')

    # settings
    for opt, val in settings.iteritems():
        if not weechat.config_is_set_plugin(opt):
            weechat.config_set_plugin(opt, val)

    if not check_database():
        say("IP database not found. You must download a database with '/country update' before "
                "using this script.")

    if not pytz_module and get_config_boolean('show_localtime'):
        error(
            "pytz module isn't installed, local time information is DISABLED. "
            "Get it from http://pytz.sourceforge.net or from your distro packages "
            "(python-tz in Ubuntu/Debian).")
        weechat.config_set_plugin('show_localtime', 'off')

    # -------------------------------------------------------------------------
    # Debug

    if weechat.config_get_plugin('debug'):
        try:
            # custom debug module I use, allows me to inspect script's objects.
            import pybuffer
            debug = pybuffer.debugBuffer(globals(), '%s_debug' % SCRIPT_NAME)
        except:
            def debug(s, *args):
                if not isinstance(s, basestring):
                    s = str(s)
                if args:
                    s = s %args
                prnt('', '%s\t%s' % (script_nick, s))
    else:
        def debug(*args):
            pass

# vim:set shiftwidth=4 tabstop=4 softtabstop=4 expandtab textwidth=100:



# -*- coding: utf-8 -*-
#
# Copyright (C) 2009-2014 Sébastien Helleu <flashcode@flashtux.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# Item with highlight/private messages.
#
# After loading this script, you can add item "hlpv" to your status
# bar with command:
#   /set weechat.bar.status.items [+tab]
#   then complete string by adding for example (without quotes): ",[hlpv]"
#
# History:
#
# 2014-05-10, Sébastien Helleu <flashcode@flashtux.org>
#     version 0.3: change hook_print callback argument type of
#                  displayed/highlight (WeeChat >= 1.0)
# 2012-01-03, Sébastien Helleu <flashcode@flashtux.org>:
#     version 0.2: make script compatible with Python 3.x
# 2009-10-02, Sebastien Helleu <flashcode@flashtux.org>:
#     version 0.1: initial release
#

SCRIPT_NAME    = "hlpvitem"
SCRIPT_AUTHOR  = "Sebastien Helleu <flashcode@flashtux.org>"
SCRIPT_VERSION = "0.3"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC    = "Item with highlight/private messages"

import_ok = True

try:
    import weechat
except:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: http://www.weechat.org/")
    import_ok = False

# script options
hlpv_settings = {
    "show_all_buffers"      : "off",      # off = hidden buffers only
    "buffer_number"         : "on",       # display buffer number before name
    "buffer_short_name"     : "on",       # use buffer short name (if off, use full name)
    "highlight"             : "on",       # display highlights in item
    "private"               : "on",       # display privates in item
    "string_highlight"      : "",         # string displayed for highlight message (before buffer)
    "string_private"        : "",         # string displayed for private message (before buffer)
    "string_delimiter"      : " > ",      # delimiter between prefix and message
    "color_string_highlight": "",         # color for string_highlight (by default == weechat.color.status_data_highlight)
    "color_string_private"  : "",         # color for string_private (by default == weechat.color.status_data_private)
    "color_buffer_number"   : "",         # color for buffer number (by default == weechat.color.status_highlight/private)
    "color_buffer_name"     : "default",  # color for buffer name
    "color_prefix"          : "white",    # color for prefix
    "color_delimiter"       : "cyan",     # color for delimiter
    "color_message"         : "default",  # color for message
    "visible_seconds"       : "7",        # amount of seconds each message is visible
}

hlpv_messages = []

def hlpv_timer():
    weechat.hook_timer(int(weechat.config_get_plugin("visible_seconds")) * 1000, 0, 1, "hlpv_timer_cb", "")

def hlpv_timer_cb(data, remaining_calls):
    """ Called when a message must be removed from list. """
    global hlpv_messages

    if len(hlpv_messages):
        hlpv_messages.pop(0)
        weechat.bar_item_update("hlpv")
        if len(hlpv_messages) > 0:
            hlpv_timer()
    return weechat.WEECHAT_RC_OK

def hlpv_item_add(buffer, highlight, prefix, message):
    """ Add message to list of messages (will be displayed by item). """
    global hlpv_messages

    if highlight:
        color_type = weechat.config_string(weechat.config_get("weechat.color.status_data_highlight"))
        color_string_highlight = weechat.config_get_plugin("color_string_highlight")
        if color_string_highlight == "":
            color_string_highlight = color_type
        string_prefix = "%s%s" % (weechat.color(color_string_highlight),
                                  weechat.config_get_plugin("string_highlight"))
    else:
        color_type = weechat.config_string(weechat.config_get("weechat.color.status_data_private"))
        color_string_private = weechat.config_get_plugin("color_string_private")
        if color_string_private == "":
            color_string_private = color_type
        string_prefix = "%s%s" % (weechat.color(color_string_private),
                                  weechat.config_get_plugin("string_private"))
    color_delimiter = weechat.color(weechat.config_get_plugin("color_delimiter"))
    if weechat.config_get_plugin("buffer_number") == "on":
        color_buffer_number = weechat.config_get_plugin("color_buffer_number")
        if color_buffer_number == "":
            color_buffer_number = color_type
        buffer_number = "%s%s%s:" % (weechat.color(color_buffer_number),
                                     weechat.buffer_get_integer(buffer, "number"),
                                     color_delimiter)
    else:
        buffer_number = ""
    color_buffer_name = weechat.color(weechat.config_get_plugin("color_buffer_name"))
    if weechat.config_get_plugin("buffer_short_name") == "on":
        buffer_name = weechat.buffer_get_string(buffer, "short_name")
    else:
        buffer_name = weechat.buffer_get_string(buffer, "name")
    color_prefix = weechat.color(weechat.config_get_plugin("color_prefix"))
    string_delimiter = weechat.config_get_plugin("string_delimiter")
    color_message = weechat.color(weechat.config_get_plugin("color_message"))
    string = "%s%s%s%s: %s%s%s%s%s%s" % (string_prefix, buffer_number,
                                         color_buffer_name, buffer_name,
                                         color_prefix, prefix,
                                         color_delimiter, string_delimiter,
                                         color_message, message)
    if len(hlpv_messages) == 0:
        hlpv_timer()
    hlpv_messages.append(string)
    weechat.bar_item_update("hlpv")

def hlpv_print_cb(data, buffer, date, tags, displayed, highlight, prefix, message):
    """ Called when a message is printed. """
    tagslist = tags.split(",")
    show_all_buffers = weechat.config_get_plugin("show_all_buffers")
    num_displayed = weechat.buffer_get_integer(buffer, "num_displayed")
    if num_displayed == 0 or show_all_buffers == "on":
        highlight_enabled = weechat.config_get_plugin("highlight")
        private_enabled = weechat.config_get_plugin("private")
        if (int(highlight) and (highlight_enabled == "on")) or (("notify_private" in tagslist) and (private_enabled == "on")):
            hlpv_item_add(buffer, int(highlight), prefix, message)
    return weechat.WEECHAT_RC_OK

def hlpv_item_cb(data, buffer, args):
    """ Callback for building hlpv item. """
    global hlpv_messages

    if len(hlpv_messages) > 0:
        return hlpv_messages[0]
    return ""

if __name__ == "__main__" and import_ok:
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION,
                        SCRIPT_LICENSE, SCRIPT_DESC,
                        "", ""):
        # set default settings
        for option, default_value in hlpv_settings.items():
            if not weechat.config_is_set_plugin(option):
                weechat.config_set_plugin(option, default_value)
        # new item
        weechat.bar_item_new('hlpv', 'hlpv_item_cb', '')
        # hook all printed messages
        weechat.hook_print("", "", "", 1, "hlpv_print_cb", "")
